
#include <stdio.h>
#include <string.h>

typedef unsigned char ui8_t;


#define HEADOFS 24 // HEADOFS+HEADLEN <= 64
#define HEADLEN 32 // HEADOFS+HEADLEN mod 8 = 0
#define FRAMESTART ((HEADOFS+HEADLEN)/8)

/*               10      B6      CA      11      22      96      12      F8      */      
char header[] = "0000100001101101010100111000100001000100011010010100100000011111";
char buf[HEADLEN+1] = "x";
int bufpos = -1;

#define FRAME_LEN 320
ui8_t frame[FRAME_LEN] = { 0x10, 0xB6, 0xCA, 0x11, 0x22, 0x96, 0x12, 0xF8};


#define MASK_LEN 64
ui8_t mask[MASK_LEN] = { 0x96, 0x83, 0x3E, 0x51, 0xB1, 0x49, 0x08, 0x98, 
                         0x32, 0x05, 0x59, 0x0E, 0xF9, 0x44, 0xC6, 0x26,
                         0x21, 0x60, 0xC2, 0xEA, 0x79, 0x5D, 0x6D, 0xA1,
                         0x54, 0x69, 0x47, 0x0C, 0xDC, 0xE8, 0x5C, 0xF1,
                         0xF7, 0x76, 0x82, 0x7F, 0x07, 0x99, 0xA2, 0x2C,
                         0x93, 0x7C, 0x30, 0x63, 0xF5, 0x10, 0x2E, 0x61,
                         0xD0, 0xBC, 0xB4, 0xB6, 0x06, 0xAA, 0xF4, 0x23,
                         0x78, 0x6E, 0x3B, 0xAE, 0xBF, 0x7B, 0x4C, 0xC1};
/* LFSR: ab i=8 (mod 64):
 * m[16+i] = m[i] ^ m[i+2] ^ m[i+4] ^ m[i+6]
 * ________________3205590EF944C6262160C2EA795D6DA15469470CDCE85CF1
 * F776827F0799A22C937C3063F5102E61D0BCB4B606AAF423786E3BAEBF7B4CC196833E51B1490898
 */

/* ------------------------------------------------------------------------------------ */

#define BAUD_RATE 4800

int sample_rate = 0, bits_sample = 0, channels = 0;
float samples_per_bit = 0;

int findstr(char *buff, char *str, long pos) {
    int i;
    for (i = 0; i < 4; i++) {
        if (buff[(pos+i)%4] != str[i]) break;
    }
    return i;
}

int read_wav_header(FILE *fp, int *sr, int *bs, int *ch) {
    char txt[5] = "\0\0\0\0";
    char buff[4];
    int byte, num, i;
    long p = 0, pos_fmt, pos_dat;
    char fmt_[5] = "fmt ";
    char data[5] = "data";

    if (fseek(fp, 0L, SEEK_SET)) return -1;
    if (fread(txt, 1, 4, fp) < 4) return -1;
    if (strncmp(txt, "RIFF", 4)) return -1;
    if (fseek(fp, 8L, SEEK_SET)) return -1;
    if (fread(txt, 1, 4, fp) < 4) return -1;
    if (strncmp(txt, "WAVE", 4)) return -1;

    // pos_fmt = 12L
    for ( ; ; ) {
        if ( (byte=fgetc(fp)) == EOF ) return -1;
        buff[p % 4] = byte;
        p++;
        if (findstr(buff, fmt_, p) == 4) break;
    }
    pos_fmt = ftell(fp)-4; //fprintf(stderr, "pos_fmt %ld\n", pos_fmt);
    if (fseek(fp, pos_fmt+10, SEEK_SET)) return -1;
    num = 0;
    for (i = 0; i < 2; i++) {
        if ( (byte=fgetc(fp)) == EOF ) return -1;
        num |= (byte << (8*i));
    }
    *ch = num;

    // if (fseek(fp, pos_fmt+12L, SEEK_SET)) return -1;  // 24L
    num = 0;
    for (i = 0; i < 4; i++) {
        if ( (byte=fgetc(fp)) == EOF ) return -1;
        num |= (byte << (8*i));
    }
    *sr = num;

    if (fseek(fp, pos_fmt+22, SEEK_SET)) return -1;  // 34L
    num = 0;
    for (i = 0; i < 2; i++) {
        if ( (byte=fgetc(fp)) == EOF ) return -1;
        num |= (byte << (8*i));
    }
    *bs = num;

    // pos_dat = 36L
    for ( ; ; ) {
        if ( (byte=fgetc(fp)) == EOF ) return -1;
        buff[p % 4] = byte;
        p++;
        if (findstr(buff, data, p) == 4) break;
    }
    pos_dat = ftell(fp)-4; //fprintf(stderr, "pos_dat %ld\n", pos_dat);
    /*
    if (fseek(fp, 36L, SEEK_SET)) return -1;
    if (fread(txt, 1, 4, fp) < 4) return -1;
    if (strncmp(txt, "data", 4)) return -1;
    */
    if (fseek(fp, pos_dat+8, SEEK_SET)) return -1;  // 44L


    fprintf(stderr, "sample_rate: %d\n", sample_rate);
    fprintf(stderr, "bits       : %d\n", bits_sample);
    fprintf(stderr, "channels   : %d\n", channels);

    if ((bits_sample != 8) && (bits_sample != 16)) return -1;

    samples_per_bit = sample_rate/(float)BAUD_RATE;

    fprintf(stderr, "samples/bit: %.2f\n", samples_per_bit);

    return 0;
}

int read_sample(FILE *fp) {  // int = i32_t
    int byte, i, ret;        // return ui16_t/ui8_t oder EOF

    for (i = 0; i < channels; i++) {
                          // i = 0: links bzw. mono
        byte = fgetc(fp);
        if (byte == EOF) return EOF;
        if (i == 0) ret = byte;
    
        if (bits_sample == 16) {
            byte = fgetc(fp);
            if (byte == EOF) return EOF;
            if (i == 0) ret +=  byte << 8;
        }

    }
                 // unsigned 8/16 bit sample >= 0
    return ret;  // EOF < 0
}

int sign(int sample) {
    int sgn = 0;
    if (bits_sample == 8) {                         // unsigned char:
        if (sample & 0x80) sgn = 1; else sgn = -1;  // 00..7F - , 80..FF: +
    }
    else if (bits_sample == 16) {
        if (sample & 0x8000) sgn = -1; else sgn = 1;
    }
    return sgn;
}

int par=1, par_alt=1;
unsigned long sample_count = 0;

int read_bits_fsk(FILE *fp, int *bit, int *len) {
    int n, sample;
    float l;

    n = 0;
    do{
        sample = read_sample(fp);  // unsigned sample;
        if (sample == EOF) return EOF; // usample >= 0
        par_alt = par;
        par = sign(sample);
        sample_count++;
        n++;
    } while (par*par_alt > 0);

    l = (float)n / samples_per_bit;  // abw = n % samples_per_bit;

    *len = (int)(l+0.5);
    *bit = (1-par_alt)/2;  // unten 1, oben -1
                           // inverse: *bit = (1+par_alt)/2;
    /* Y-offset ? */

    return 0;
}

/* ------------------------------------------------------------------------------------ */


int bits2byte(char bits[]) {
    int i, byteval=0, d=1;
    for (i = 0; i < 8; i++) {     // little endian
    /* for (i = 7; i >= 0; i--) { // big endian */
        if      (bits[i] == 1)  byteval += d;
        else if (bits[i] == 0)  byteval += 0;
        else return 0x100;
        d <<= 1;      
    }
    return byteval;
}


void inc_bufpos() {
  bufpos = (bufpos+1) % HEADLEN;
}

int compare() {
    int i=0, j = bufpos;

    while (i < HEADLEN) {
        if (j < 0) j = HEADLEN-1;
        if (buf[j] != header[HEADLEN+HEADOFS-1-i]) break;
        j--;
        i++;
    }
    return i;
}

ui8_t xorbyte(int pos) {
    return  frame[pos] ^ mask[pos % MASK_LEN];
}

/*
#include "rs41pos.c"
void print_frame(int len) {
    print_position();
}
*/
void print_frame(int len) {
    int i;
    ui8_t byte;
    for (i = 0; i < len; i++) {
        //byte = frame[i];
        byte = xorbyte(i);
        fprintf(stdout, "%02x", byte);
    }
    fprintf(stdout, "\n");
}


int main(int argc, char *argv[]) {

    FILE *fp;
    char bitbuf[8];
    int bit_count = 0,
        byte_count = FRAMESTART,
        header_found = 0,
        byte, i;
    int bit, len;

    argv++;
            if ( !(fp = fopen(*argv, "rb")) ) {
                fprintf(stderr, "%s konnte nicht geoeffnet werden\n", *argv);
                return -1;
            }

    i = read_wav_header(fp, &sample_rate, &bits_sample, &channels);
    if (i) {
        fclose(fp);
        return -1;
    }


    while (!read_bits_fsk(fp, &bit, &len)) {

        if (len == 0) { // reset_frame();
            if (byte_count > FRAME_LEN-20) print_frame(byte_count);
            bit_count = 0;
            byte_count = FRAMESTART;
            header_found = 0;
            inc_bufpos();
            buf[bufpos] = 'x';
            continue;   // ...
        }

        for (i = 0; i < len; i++) {

            inc_bufpos();
            buf[bufpos] = 0x30 + bit;  // Ascii

            if (!header_found) {
                if (compare() >= HEADLEN) header_found = 1;
            }
            else {
                bitbuf[bit_count] = bit;
                bit_count++;
            
                if (bit_count == 8) {
                    bit_count = 0;
                    byte = bits2byte(bitbuf);
                    frame[byte_count] = byte;
                    byte_count++;
                    if (byte_count == FRAME_LEN) {
                        byte_count = FRAMESTART;
                        header_found = 0;
                        print_frame(FRAME_LEN);
                    }
                }
            }

        }

    }
    fclose(fp);

    return 0;
}


